#!/usr/bin/env python
import os
import csv
import sys
import argparse
import json
import random
from twython import Twython
from twilight import stderrn

# import tweepy
# auth = tweepy.auth.OAuthHandler(consumer_key, consumer_secret)
# auth.set_access_token(access_key, access_pass)
# api = tweepy.API(auth)

# tweets = {} #Maps IDs to Tweet Objects
# for tweet_id in ids_list:
#     tweets[tweet_id] = api.get_status(tweet_id)


def main():
    parser = argparse.ArgumentParser(description='Crawl all tweets from a given user.')
    parser.add_argument('--output', default='-', help='output file (defaults to STDOUT)')
    # parser.add_argument('--consumer_key', help='Twitter user to use to crawl')
    # parser.add_argument('--password', help='Password for crawl connections.')
    parser.add_argument('--screen_name', help='Twitter user to crawl')
    opts = parser.parse_args()

    accounts = [row for row in csv.DictReader(open(os.path.expanduser('~/.twitter')))]
    account = random.sample(accounts, 1)[0]
    client = Twython(
        account['consumer_key'], account['consumer_secret'],
        account['access_token'], account['access_token_secret'])

    # auth = (opts.username, opts.password)
    params = dict(
        screen_name=opts.screen_name,
        include_entities='true',
        contributor_details='true',
        include_rts='true',
        count=200)

    counts = []
    output = sys.stdout if opts.output == '-' else open(opts.output, 'a')

    while True:
        tweets = client.get_user_timeline(**params)
        count = len(tweets)
        for tweet in tweets:
            json.dump(tweet, output)
            output.write('\n')

        counts.append(count)
        if count > 0:
            try:
                params['max_id'] = tweets[-1]['id'] - 1
            except Exception, exc:
                print exc, tweets
                return

        empties = len([count for count in counts if count < 10])
        if empties > 5:
            break

    stderrn('Done downloading %d tweets from %s.' % (sum(counts), opts.screen_name))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import sys
import time
import subprocess
import ujson as json  # ujson returns python None for json null


def stderr(s):
    sys.stderr.write(s)
    sys.stderr.flush()

def stderrn(s=''):
    stderr('%s\n' % s)

def stdout(s):
    sys.stdout.write(s)
    sys.stdout.flush()

def stdoutn(s=''):
    stdout('%s\n' % s)

def stdout_tabs(*args):
    stdout('\t'.join(args))
    stdout('\t')

def sh(raw_args):
    args = map(str, raw_args)
    print '$ %s' % ' '.join(args)
    return subprocess.check_output(args)

def flatten_dict(d, prefix=None):
    prefix = prefix or []
    for k, v in d.items():
        if isinstance(v, dict):
            for keyval in flatten_dict(v, prefix + [k]):
                yield keyval
        else:
            yield ('.'.join(prefix + [k]), v)


class Tweet(object):
    translations = {ord('\t'): u' ', ord('\n'): u' ', ord('\r'): u''}

    @classmethod
    def parse_js_date(cls, s):
        return time.strptime(s, '%a %b %d %H:%M:%S +0000 %Y')

    @classmethod
    def date_to_compact(cls, dt):
        return time.strftime('%Y%m%dT%H%M%S', dt)

    def __init__(self, values):
        self.values = values

    @classmethod
    def from_json(cls, tweet_json_string):
        return cls.from_dict(json.loads(tweet_json_string))

    @classmethod
    def from_tsv(cls, line):
        return cls(line.split(u'\t'))

    def to_tsv(self):
        return u'\t'.join(self.values)

    def get(self, key):
        return self.values[self.cols.index(key)]

    def set(self, key, val):
        self.values[self.cols.index(key)] = val

    @property
    def is_arabic(self):
        # more than half arabic = is arabic
        text = self.get('text')
        return len([ch for ch in text if 1536 <= ord(ch) <= 1791]) > (len(text) / 2)

    def text_rtl_to_ltr(self):
        # if arabic were capitals: abcdefZYXUlokpTRQNML
        #         it would become: abcdefUXYZlokpLMNQRT
        chars = list(self.get('text'))
        span_start = None
        span_end = None
        for i, ch in enumerate(chars):
            chord = ord(ch)
            if 1536 <= chord <= 1791:
                # arabic char!
                if not span_start:
                    span_start = i
                span_end = i
            elif not chord == 32 and span_start:
                # it's not a space, and we ARE in an Arabic span.
                chars[span_start:span_end+1] = chars[span_end:span_start-1:-1]
                span_start = None

        return u''.join(chars)


def map_stdin(fn):
    """
    Streams stdin line-by-line through the given `fn`, directly to stdout.
      each line will be utf-8 decoded, such that fn will get a unicode string
    signature of fn: `def fn(unicode_line):`
    The return value of fn is discarded
    When `fn` raises an error besides KeyboardInterrupt, or EOFError, this
        method will write the error to stderr and continue with the next line.
    It does not fuck with stdout, but will log errors to stderr
    If KeyboardInterrupt, EOFError are seen, the loop will exit.
    """
    line = sys.stdin.readline()
    while line:
        try:
            fn(line.decode('utf-8'))
        except (KeyError, ValueError), exc:
            stderr(u'%s\n  --> %s' % (exc, line.decode('utf-8')))
        except (KeyboardInterrupt, EOFError):
            stderr(u'\n  -- Interrupted, exiting --\n')
            break

        line = sys.stdin.readline()

    stderr('  -- Done --')


class Binner(object):
    # Each line of output will be just a datetime/key and a bunch of texts.
    def __init__(self, key_fn, val_fn):
        self.key_fn = key_fn
        self.val_fn = val_fn
        self.buffer = []
        self.key = None

    def add(self, tweet):
        """
        The idea is that I'll take a gzipped (.gz) list of flattened tweets (.ttv1),
        e.g. syria_20120615-20120829.ttv1.gz, and ...
        $ zcat syria_*.ttv1.gz | tweetop bin 6h > syria_20120906.texts
        """
        new_key, new_val = self.key_fn(tweet), self.val_fn(tweet)
        if new_key != self.key:
            if self.key:
                self.flush()
            self.key = new_key
        self.buffer.append(new_val)

    def flush(self):
        line = u'%s\t%s' % (self.key, u' '.join(self.buffer))
        self.buffer = []
        stdout(line.encode('utf-8'))



#!/usr/bin/env python
import sys
import argparse
from twilight import stderr, stdout

def main():
    parser = argparse.ArgumentParser(description='''
        Search ttv2 files for matches in a certain bounding box.
        geogrep expects raw ttv2 on STDIN.''')
    parser.add_argument('north', type=float)
    parser.add_argument('east', type=float)
    parser.add_argument('south', type=float)
    parser.add_argument('west', type=float)
    opts = parser.parse_args()
    north, east, south, west = opts.north, opts.east, opts.south, opts.west

    matches = 0
    total = 0

    for line in sys.stdin:
        lon_lat = line.split('\t', 5)[3]
        if lon_lat != '':
            lon, lat = map(float, lon_lat.split(','))
            if (opts.north >= lat >= opts.south) and (opts.east >= lon >= opts.west):
                matches += 1
                stdout(line)
        total += 1
        if total % 1000 == 0:
            stderr('\r%.2f%% of %d' % ((matches * 100.0) / total, total))

if __name__ == '__main__':
    main()



import shapefile

sf = shapefile.Reader('/home/chbrown/data/borders/thematicmap-simple')
fields = [f[0] for f in sf.fields[1:]]
# records[245] = {'AREA': 0,
#  'FIPS': 'TW',
#  'ISO2': 'TW',
#  'ISO3': 'TWN',
#  'LAT': 23.754,
#  'LON': 120.946,
#  'NAME': 'Taiwan',
#  'POP2005': 0,
#  'REGION': 0,
#  'SUBREGION': 0,
#  'UN': 158}]


def polygon_contains(poly, x, y):
    '''Thanks, http://www.ariel.com.au/a/python-point-int-poly.html'''
    n = len(poly)
    inside = False

    p1x, p1y = poly[0]
    for i in range(n + 1):
        p2x, p2y = poly[i % n]
        if y > min(p1y, p2y):
            if y <= max(p1y, p2y):
                if x <= max(p1x, p2x):
                    if p1y != p2y:
                        xinters = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                    if p1x == p2x or x <= xinters:
                        inside = not inside
        p1x, p1y = p2x, p2y

    return inside


def mean(floats):
    return sum(floats)/len(floats)


def first_country_containing(tweet):
    coords = tweet['coordinates']
    if coords:
        if coords['type'] == u'Point':
            lon, lat = coords['coordinates']
        elif coords['type'] == u'Polygon':
            # u'coordinates': [[[-7.6028565, 33.4087762], [-7.4503
            lons, lats = zip(*coords['coordinates'][0])
            lon, lat = mean(lons), mean(lats)
        else:
            lon, lat = -999, -999

        for country in countries:
            if country.bbox_contains(lon, lat):
                if country.polygons_contain(lon, lat):
                    return country
    return None


class Country(object):
    def __init__(self, record, shape):
        self.sw_lon, self.sw_lat, self.ne_lon, self.ne_lat = shape.bbox
        parts = shape.parts.tolist() + [len(shape.points)]
        self.polygons = [shape.points[i:j] for i, j in zip(parts, parts[1:])]
        for field, val in zip(fields, record):
            setattr(self, field, val)

    def bbox_contains(self, lon, lat):
         # or self.sw_lon > lon > self.ne_lon
        if self.sw_lon < lon < self.ne_lon:
            #  or self.sw_lat > lat > self.ne_lat
            if self.sw_lat < lat < self.ne_lat:
                return True
        return False

    def polygons_contain(self, lon, lat):
        return any(polygon_contains(poly, lon, lat) for poly in self.polygons)

    def __str__(self):
        return '%s (%s/%s) SW: [%0.7f, %0.7f] NE: [%0.7f, %0.7f]' % (self.NAME, self.ISO2, self.ISO3,
            self.sw_lon, self.sw_lat, self.ne_lon, self.ne_lat)

countries = [Country(record, shape) for record, shape in zip(sf.records(), sf.shapes())]



#!/usr/bin/env python
import sys
sys.path.append('..')
from flatten import colnames

text_col = colnames.index('text')

counter = 0
raw = 'START'
while raw:
    try:
        values = raw.decode('utf-8').split(u'\t')
        tweet_text = values[text_col]
        char_codes = map(ord, tweet_text)

        arabic_chars = [char_code for char_code in char_codes if 1536 <= char_code <= 1791]
        if len(arabic_chars) > (len(char_codes) / 2.0):
            sys.stdout.write('%s\n' % tweet_text.encode('utf-8'))

        if counter % 10000 == 0:
            sys.stderr.write('\r  #%10d   ' % counter)
            sys.stderr.flush()
        counter += 1
    except Exception, exc:
        print exc

    raw = sys.stdin.readline()

sys.stderr.write('\n%d\n\n' % counter)
sys.stdout.flush()



#!/usr/bin/env python
import os
import re
import gzip
import bz2
import argparse
import psutil
from twilight import stdout_tabs, stdoutn, stderrn
from twilight.ttv import TTV2


description = '''Convert json.gz files to ttv2.
    ttv2izer looks at /usr/local/data/twitter/*.json.gz files and sequentially ttv2'izes them.
    To check if it's behaving:
    cat yourfile.bz2 | awk 'BEGIN{FS="\\t"}{print NF}' # should only output 26's'''


def get_openfiles():
    openfiles = []
    for p in psutil.get_process_list():
        try:
            openfiles += [os.path.basename(openfile.path) for openfile in p.get_open_files()]
        except psutil.AccessDenied:
            pass
    return openfiles


def open_any(path):
    '''Open any Tweets raw json file, depending on its extension.'''
    if path.endswith('gz'):
        return gzip.GzipFile(path)
    elif path.endswith('bz2'):
        return bz2.BZ2File(path)
    else:
        return open(path)


def convert_json_to_ttv2(json_path, ttv2_path):
    # the given json_path can be either gzip'ed or bzip2'ed or not
    # the resulting file at ttv2_path will always be bzip2'ed
    with bz2.BZ2File(ttv2_path, 'w') as ttv2_fp:
        with open_any(json_path) as json_fp:
            for line in json_fp:
                try:
                    line = line.decode('utf-8')
                    tweet = TTV2.from_json(line)
                    ttv2_fp.write(tweet.to_tsv().encode('utf-8'))
                    ttv2_fp.write('\n')
                except UnicodeDecodeError, exc:
                    stderrn("Error decoding string: %s (%s)" % (line.strip(), exc))
                except Exception, exc:
                    stderrn("Error translating json: %s (%s)" % (line.strip(), exc))


def main():
    parser = argparse.ArgumentParser(description=description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--directory', default='/usr/local/data/twitter')
    parser.add_argument('--delete', action='store_true')
    parser.add_argument('--overwrite', action='store_true')
    opts = parser.parse_args()

    os.chdir(opts.directory)
    # filter out the openfiles, and only look at .json or .json.gz files
    openfiles = get_openfiles()
    filenames = [filename for filename in os.listdir('.') if re.search('.json(.gz)?$', filename) and filename not in openfiles]
    for filename in filenames:
        ttv2_filename = '%s.ttv2.bz2' % filename.split('.')[0]

        stdout_tabs(filename, '>', ttv2_filename)

        if not os.path.exists(ttv2_filename) or opts.overwrite:
            stdout_tabs('ttv2izing')
            convert_json_to_ttv2(filename, ttv2_filename)
        else:
            stdout_tabs('ttv2ized')

        fraction = os.stat(ttv2_filename).st_size / float(os.stat(filename).st_size)
        stdout_tabs('%0.2f%% as big' % (fraction * 100))

        deletable = False
        if filename.endswith('.gz') and 0.18 < fraction < 0.5:
            deletable = True
        elif 0.02 < fraction < 0.06:
            deletable = True

        if deletable and opts.delete:
            stdout_tabs('rm')
            os.remove(filename)
        elif deletable:
            stdout_tabs('deletable')
        else:
            stdout_tabs('bad-size')

        stdoutn()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import csv
import argparse
from itertools import groupby

parser = argparse.ArgumentParser(description='Common Twitter Operations')
parser.add_argument('csvfile', type=str)
parser.add_argument('--key', type=str)
parser.add_argument('--out', type=str)
# parser.add_argument('--include', nargs='*')
# parser.add_argument('--exclude', nargs='*')

opts = parser.parse_args()

with open(opts.csvfile) as fp:
    reader = csv.DictReader(fp)
    rows = list(reader)
    cols = reader.fieldnames

if opts.key == 'superfile':
    superfile_groups = [
        ['libya', 'libyan', 'gaddafi-ar', 'gaddafi', 'benghazi', 'libya-ar', 'benghazi-ar'],
        ['egypt', 'cairo', 'egypt-ar', 'cairo-ar'],
        ['gop', 'romney'],
        ['dem', 'hillary-clinton', 'barack-obama'],
        ['yemen', 'yemen-ar', 'sana', 'sana-ar'],
        ['arab-spring', 'middle-east'],
        ['anti-islam', 'morris'],
        ['embassy', 'christopher-stevens', 'american-ambassador', 'ambassador', 'consulate']]
    for row in rows:
        for superfile_group in superfile_groups:
            if row['file'] in superfile_group:
                row['superfile'] = superfile_group[0]
    def key_func(row):
        # print row
        return (row['superfile'], row['key'])

aggregate_rows = []
for group_key, group_rows in groupby(sorted(rows, key=key_func), key_func):
    superfile, dt = group_key
    group_rows = list(group_rows)
    aggregate_row = dict(superfile=superfile, dt_dummy=dt)
    for col in cols:
        try:
            aggregate_row[col] = sum(row[col] for row in group_rows)
        except:
            aggregate_row[col] = group_rows[0][col]

    aggregate_rows.append(aggregate_row)


cols += ['superfile', 'dt_dummy']

with open(opts.out, 'w') as fp:
    writer = csv.DictWriter(fp, cols)
    writer.writeheader()
    writer.writerows(aggregate_rows)


# eg: aggregate libya-egypt-12h.csv --key superfile --out libya-egypt-12h-super.csv



#!/usr/bin/env python
import re
import sys
import argparse
import ujson as json
from twitter import map_stdin, stdout, stderr, Binner, TTV2, flatten_dict

def main():
    parser = argparse.ArgumentParser(description='Common Twitter Operations')
    parser.add_argument('op', type=str)
    parser.add_argument('--unit', type=str, help="Use with --op bin")
    parser.add_argument('--count', type=float, help="Use with --op bin")
    parser.add_argument('--files', nargs='*', help="Use with --op glomcsv")
    opts = parser.parse_args()

    if opts.op == 'json2ttv2':
        def mapper(line):
            tweet = TTV2.from_json(line)
            stdout(tweet.to_tsv().encode('utf-8'))
        map_stdin(mapper)
    elif opts.op == 'bin':
        from datetime import datetime
        val_func = lambda t: t.text_rtl_to_ltr()
        if opts.unit == 'm':
            # YYYYMMDD:HHMM = created_at[:13]
            key_func = lambda t: t.t.get('created_at')[:13]
        elif opts.unit == 'h':
            # YYYYMMDD:HH = created_at[:11]
            h = opts.count
            def key_func(t):
                dt = datetime.strptime(t.get('created_at'), '%Y%m%dT%H%M%S')
                hours = int((dt.hour + (dt.minute / 60.0))/h)*h + (h/2)
                return '%sT%02d0000' % (dt.strftime('%Y%m%d'), hours)
        elif opts.unit == 'd':
            # YYYYMMDD = created_at[:8]
            key_func = lambda t: t.get('created_at')[:8]

        binner = Binner(key_func, val_func)
        def mapper(line):
            tweet = TTV2.from_tsv(line)
            binner.add(tweet)
        map_stdin(mapper)

    elif opts.op == 'sa':
        from lexicons import arabsenti, finn, liwc

        # u"\"!.:;,/\\#+"
        strip = dict((ord(ch), u' ') for ch in u'!"#%\'()*+,-./:;<=>?@[\]^_`{|}~')

        def mapper(line):
            try:
                key, text = line.split(u'\t', 1)
                at_mentions = re.findall(r'@\w{4,}', text)
                tokens = text.lower().translate(strip).split()

                stderr('key: %s, %d chars, %d words' % (key, len(text), len(tokens)))

                sentiments = dict(finn=finn.tokens(tokens),
                    arabsenti=arabsenti.tokens(tokens),
                    liwc=liwc.tokens(tokens),
                    wc=len(tokens),
                    chars=len(text),
                    at_mentions=len(at_mentions))

                stdout('%s\t%s' % (key, json.dumps(sentiments)))
            except:
                stderr('%s appears to have no key!' % line[:10000])

        map_stdin(mapper)

        # cat ~/Desktop/syria-snippet.keytext.gz | gunzip | cut -c -100000 | ./sa
        # cat ~/Desktop/syria-snippet.keytext.gz | gunzip | ./sa
        # | gzip > ~/Desktop/syria-snippet.keysenti.gz

    elif opts.op == 'json2csv':
        import csv
        # cat syria_20120615-20120829.sentiments | cli json2csv > syria_20120615-20120829.csv
        rows = []
        cols = ['key']
        for line in sys.stdin:
            tabs = line.split('\t')
            json_dict = json.loads(tabs[-1])
            keyvals = dict(flatten_dict(json_dict))
            keyvals['key'] = '-'.join(tabs[0:-1])
            rows.append(keyvals)
            cols = cols + list(set(keyvals.keys()) - set(cols))

        cols = cols[0:1] + sorted(cols[1:])
        writer = csv.DictWriter(sys.stdout, cols, restval='', extrasaction='raise', dialect=csv.excel)
        writer.writeheader()
        for row in rows:
            writer.writerow(row)

    elif opts.op == 'glomcsv':
        filenames = opts.files

        initialized = False
        for filename in filenames:
            basename = filename.replace('.csv', '')
            with open(filename) as fp:
                headers = fp.readline().strip()
                if not initialized:
                    stdout('file,%s' % headers)
                    initialized = True
                for line in fp:
                    stdout('%s,%s' % (basename, line.strip()))


    elif opts.op == 'geo':
        from twitter.geo import first_country_containing
        # call like:
        # $ cat /usr/local/data/twitter/nafrica_2012-06* | gunzip | ./geofilter | gzip > ~/data/nafrica_arabic_2012-06.tsv.gz

        def arabic_countries(tweet):
            if tweet.is_arabic:
                country = first_country_containing(tweet)
                if country:
                    tsv = u'\t'.join([country.ISO3] + tweet.flatten())
                    stdout(tsv.encode('utf-8'))
                else:
                    stderr('No coord info!')

        def arabic(tweet):
            if tweet.is_arabic:
                stdout(tweet.tsv().encode('utf-8'))

        def everything(tweet):
            stdout(tweet.tsv().encode('utf-8'))

        map_stdin(everything)
    else:
        print "That command line call wasn't recognized"

if __name__ == '__main__':
    main()



from setuptools import setup
import json
from os import path
here = path.abspath(path.dirname(__file__))
package = json.load(open(path.join(here, 'package.json')))

setup(
    name=str(package['name']),
    version=str(package['version']),
    author='Christopher Brown',
    author_email='chrisbrown@utexas.edu',
    packages=[package['name']],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'pyshp',
        'psutil',
        'twython',
    ],
    entry_points={
        'console_scripts': [
            'json2ttv2 = twilight.json2ttv2:main',
            'tweetop = twilight.tweetop:main',
            'geogrep = twilight.geogrep:main',
            'twitter-user = twilight.user:main',
        ],
    },
)



from twilight import Tweet

class TTV1(Tweet):
    cols = ['id_str', 'created_at', 'text', 'retweeted', 'retweet_count', 'coordinates', 'geo', 'user_id_str', 'user_created_at', 'user_screen_name', 'user_name', 'user_location', 'user_url', 'user_statuses_count', 'user_followers_count', 'user_friends_count', 'user_geo_enabled', 'user_time_zone', 'user_lang', 'user_utc_offset']

    @classmethod
    def parse_date(cls, s):
        if '+0000' in s:
            return cls.date_to_compact(cls.parse_js_date(s))
        return s

    @classmethod
    def from_dict(cls, store):
        user = store['user']
        coordinates = store['coordinates']
        coords = u''
        if coordinates and coordinates['type'] == u'Point':
            coords = u','.join(map(unicode, coordinates['coordinates']))
        values = (
            store['id_str'],
            cls.parse_date(store['created_at']),
            store['text'].translate(cls.translations).replace('&amp;', '&').replace('&quot;', '"').replace('&apos;', "'"),
            u'T' if store['retweeted'] else u'F',
            unicode(store['retweet_count']),
            coords,
            user['id_str'],
            cls.parse_date(user['created_at']),
            user['screen_name'],
            user['name'].translate(cls.translations),
            user['location'].translate(cls.translations),
            user['url'].translate(cls.translations) if user['url'] else '',
            str(user['statuses_count']),
            str(user['followers_count']),
            str(user['friends_count']),
            u'T' if user['geo_enabled'] else u'F',
            user['time_zone'] or '',
            user['lang'],
            str(user['utc_offset'] or '')
        )
        return cls(values)

class TTV2(Tweet):
    cols = ['id', 'created_at', 'text', 'coordinates', 'place_id', 'place_str',
        'in_reply_to_status_id', 'in_reply_to_screen_name', 'retweet_id',
        'retweet_count', 'user_screen_name', 'user_id', 'user_created_at',
        'user_name', 'user_description', 'user_location', 'user_url',
        'user_statuses_count', 'user_followers_count', 'user_friends_count',
        'user_favourites_count', 'user_geo_enabled', 'user_default_profile',
        'user_time_zone', 'user_lang', 'user_utc_offset']

    # ttv2 is the tweet tab-separated format version 2:
    # the spec is:
    #   0: tweet_id
    #   1: created_at parsed into YYYYMMDDTHHMMSS, implicitly UTC
    #   2: text, newlines and tabs converted to spaces, html entities replaced, t.co urls resolved
    #   3: lon,lat
    #   4: place_id
    #   5: place_str
    #   6: in_reply_to_status_id
    #   7: in_reply_to_screen_name
    #   8: retweet_id id of the original tweet
    #   9: retweet_count
    #  10: user.screen_name
    #  11: user.id
    #  12: user.created_at parsed into YYYYMMDDTHHMMSS
    #  13: user.name
    #  14: user.description
    #  15: user.location
    #  16: user.url
    #  17: user.statuses_count
    #  18: user.followers_count
    #  19: user.friends_count
    #  20: user.favourites_count
    #  21: user.geo_enabled
    #  22: user.default_profile
    #  23: user.time_zone
    #  24: user.lang
    #  25: user.utc_offset

    @classmethod
    def from_dict(cls, d):
        created_at = cls.date_to_compact(cls.parse_js_date(d['created_at']))

        text = d['text'].translate(cls.translations)
        for url in d['entities']['urls']:
            text = text.replace(url['url'], url['expanded_url'] or url['url'])
        for url in d['entities'].get('media', []):
            text = text.replace(url['url'], url['media_url'] or url['url'])
        text = text.\
            replace('&amp;', '&').replace('&quot;', '"').replace('&apos;', "'").\
            replace('&gt;', '>').replace('&lt;', '<')

        coords = '%.8f,%.8f' % tuple(d['coordinates']['coordinates']) if d['coordinates'] else ''
        place_id = ''
        place_str = ''
        if d['place']:
            place_id = d['place']['id']
            if d['place']['full_name'] == d['place']['country']:
                place_str = d['place']['full_name']
            else:
                place_str = '%s; %s' % (d['place']['full_name'], d['place']['country'])

        retweet_id = d['retweeted_status']['id_str'] if 'retweeted_status' in d and d['retweeted_status'] else ''

        user = d.get('user', {})
        user_created_at = cls.date_to_compact(cls.parse_js_date(user['created_at'])) if 'created_at' in user else ''
        return cls((
            d['id_str'],
            created_at,
            text,
            coords,
            place_id,
            place_str,
            d['in_reply_to_status_id_str'] or u'',
            d['in_reply_to_screen_name'] or u'',
            retweet_id,
            unicode(d.get('retweet_count', '')),
            user.get('screen_name', ''),
            user.get('id_str', ''),
            user_created_at,
            user.get('name', u'').translate(cls.translations),
            (user.get('description') or u'').translate(cls.translations),
            (user.get('location') or u'').translate(cls.translations),
            (user.get('url') or u'').translate(cls.translations),
            unicode(user.get('statuses_count', 0)),
            unicode(user.get('followers_count', 0)),
            unicode(user.get('friends_count', 0)),
            unicode(user.get('favourites_count', 0)),
            u'T' if user.get('geo_enabled', False) else u'F',
            u'T' if user.get('default_profile', False) else u'F',
            user.get('time_zone') or u'',
            user.get('lang') or u'',
            unicode(user.get('utc_offset') or '')
        ))
